import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from vtk import vtkPolyData  # pylint: disable=no-name-in-module

from abm_shape_collection.construct_mesh_from_coeffs import construct_mesh_from_coeffs


def construct_mesh_from_points(
    pca: PCA,
    points: np.ndarray,
    feature_names: list[str],
    order: int,
    prefix: str = "",
    suffix: str = "",
) -> vtkPolyData:
    """
    Construct mesh given PCA transformation points.

    Parameters
    ----------
    pca
        Fit PCA object.
    points
        Select point in PC space.
    feature_names
        Spherical harmonics coefficient names.
    order
        Order of the spherical harmonics coefficient parametrization.
    prefix
        Prefix string for all coefficient columns.
    suffix
        Suffix string for all coefficient columns.

    Returns
    -------
    :
        Mesh object.
    """

    coeffs = pd.Series(pca.inverse_transform(points), index=feature_names)
    return construct_mesh_from_coeffs(coeffs, order, prefix, suffix)



import numpy as np


def extract_voxel_contours(
    all_voxels: list[tuple[int, int, int]],
    projection: str,
    box: tuple[int, int, int],
) -> list[list[tuple[int, int]]]:
    """
    Extract contours from list of voxels in specified projection direction.

    Parameters
    ----------
    all_voxels
        List of all voxels in region.
    projection
        Projection direction.
    box
        Bounding box.

    Returns
    -------
    :
        List of list of contour points.
    """

    voxels = set()
    length, width, height = box

    if projection == "top":
        voxels.update({(x, y) for x, y, _ in all_voxels})
        x_bounds = length
        y_bounds = width
    elif projection == "side1":
        voxels.update({(x, z) for x, _, z in all_voxels})
        x_bounds = length
        y_bounds = height
    else:
        voxels.update({(y, z) for _, y, z in all_voxels})
        x_bounds = width
        y_bounds = height

    array = np.full((x_bounds, y_bounds), fill_value=False)
    array[tuple(np.transpose(list(voxels)))] = True

    edges = get_array_edges(array)
    contours = connect_array_edges(edges)

    return [merge_contour_edges(contour) for contour in contours]


def get_array_edges(array: np.ndarray) -> list[list[tuple[int, int]]]:
    """
    Get edges of region in binary array.

    Parameters
    ----------
    array
        Binary array.

    Returns
    -------
    :
        List of unconnected region edges.
    """

    edges = []
    x, y = np.nonzero(array)

    for i, j in zip(x.tolist(), y.tolist()):
        if j == array.shape[1] - 1 or not array[i, j + 1]:
            edges.append([(i, j + 1), (i + 1, j + 1)])

        if i == array.shape[0] - 1 or not array[i + 1, j]:
            edges.append([(i + 1, j), (i + 1, j + 1)])

        if j == 0 or not array[i, j - 1]:
            edges.append([(i, j), (i + 1, j)])

        if i == 0 or not array[i - 1, j]:
            edges.append([(i, j), (i, j + 1)])

    return edges


def connect_array_edges(edges: list[list[tuple[int, int]]]) -> list[list[tuple[int, int]]]:
    """
    Group unconnected region edges into connected contour edges.

    Parameters
    ----------
    edges
        List of unconnected region edges.

    Returns
    -------
    :
        List of connected contour edges.
    """

    contours: list[list[tuple[int, int]]] = []

    while edges:
        contour = edges[0]
        contour_length = 0
        edges.remove(contour)

        while contour_length != len(contour):
            contour_length = len(contour)

            forward = list(filter(lambda edge: contour[-1] == edge[0], edges))

            if len(forward) > 0:
                edges.remove(forward[0])
                contour.extend(forward[0][1:])

            backward = list(filter(lambda edge: contour[-1] == edge[-1], edges))

            if len(backward) > 0:
                edges.remove(backward[0])
                contour.extend(list(reversed(backward[0]))[1:])

            if contour_length == len(contour):
                contours.append(list(contour))

    return sorted(contours, key=len)


def merge_contour_edges(contour: list[tuple[int, int]]) -> list[tuple[int, int]]:
    """
    Merge connected contour edges.

    Parameters
    ----------
    contour
       List of connected contour edges.

    Returns
    -------
    :
        List of connected contours.
    """

    merged = contour.copy()

    for (x0, y0), (x1, y1), (x2, y2) in zip(contour, contour[1:], contour[2:]):
        area = x0 * (y1 - y2) + x1 * (y2 - y0) + x2 * (y0 - y1)
        if area == 0:
            merged.remove((x1, y1))

    return merged



import numpy as np
from sklearn.decomposition import PCA


def fit_pca_model(features: np.ndarray, components: int, ordering: np.ndarray) -> PCA:
    """
    Fit PCA model to given data.

    Parameters
    ----------
    features
        Feature data (shape = num_samples x num_feature).
    components
        Number of components to keep.
    ordering
        Data used to reorient components.

    Returns
    -------
    :
        Fit PCA object.
    """

    # Fit data.
    pca = PCA(n_components=components)
    pca = pca.fit(features)

    # Reorient features by ordering data.
    transform = pca.transform(features)
    for i in range(components):
        pearson = np.corrcoef(ordering, transform[:, i])
        if pearson[0, 1] < 0:
            pca.components_[i] = pca.components_[i] * -1

    return pca



from typing import Callable

import numpy as np
import pandas as pd
from sklearn.decomposition import PCA

from abm_shape_collection.construct_mesh_from_points import construct_mesh_from_points
from abm_shape_collection.extract_mesh_projections import extract_mesh_projections


def extract_shape_modes(
    pca: PCA,
    data: pd.DataFrame,
    components: int,
    regions: list[str],
    order: int,
    delta: float,
    _construct_mesh_from_points: Callable = construct_mesh_from_points,
    _extract_mesh_projections: Callable = extract_mesh_projections,
) -> dict:
    """
    Extract shape modes (latent walks in PC space) at the specified intervals.

    Parameters
    ----------
    pca
        Fit PCA object.
    data
        Sample data, with shape coefficients as columns.
    components
        Number of shape coefficients components.
    regions
        List of regions.
    order
        Order of the spherical harmonics coefficient parametrization.
    delta
        Interval for latent walk, bounded by -2 and +2 standard deviations.

    Returns
    -------
    :
        Map of regions to lists of shape modes at select points.
    """

    # pylint: disable=too-many-locals

    # Transform data into shape mode space.
    columns = data.filter(like="shcoeffs").columns
    transform = pca.transform(data[columns].values)

    # Calculate transformed means and standard deviations.
    means = transform.mean(axis=0)
    stds = transform.std(axis=0, ddof=1)

    # Create bins.
    map_points = np.arange(-2, 2.5, delta)
    bin_edges = [-np.inf] + [point + delta / 2 for point in map_points[:-1]] + [np.inf]
    transform_binned = np.digitize(transform / stds, bin_edges)

    # Initialize output dictionary.
    shape_modes: dict[str, list] = {}

    for region in regions:
        region_shape_modes = []

        suffix = f".{region}" if region != "DEFAULT" else ""
        offsets = calculate_region_offsets(data, region)

        for component in range(components):
            point_vector = np.zeros(components)

            for point in map_points:
                point_bin = np.digitize(point, bin_edges)
                point_vector[component] = point

                vector = means + np.multiply(stds, point_vector)
                indices = transform_binned[:, component] == point_bin

                mesh = _construct_mesh_from_points(pca, vector, columns, order, suffix=suffix)

                if region == "DEFAULT" or not any(indices):
                    offset = None
                else:
                    offset = (
                        offsets["x"][indices].mean(),
                        offsets["y"][indices].mean(),
                        offsets["z"][indices].mean(),
                    )

                region_shape_modes.append(
                    {
                        "mode": component + 1,
                        "point": point,
                        "projections": _extract_mesh_projections(
                            mesh, extents=False, offset=offset
                        ),
                    }
                )

        shape_modes[region] = region_shape_modes

    return shape_modes


def calculate_region_offsets(data: pd.DataFrame, region: str) -> dict:
    """
    Calculate offsets for non-default regions.

    Parameters
    ----------
    data
        Centroid location data.
    region
        Name of region (skipped if region is DEFAULT).

    Returns
    -------
    :
        Map of offsets in the x, y, and z directions.
    """

    if region == "DEFAULT":
        return {}

    x_deltas = data[f"CENTER_X.{region}"].to_numpy() - data["CENTER_X"].to_numpy()
    y_deltas = data[f"CENTER_Y.{region}"].to_numpy() - data["CENTER_Y"].to_numpy()
    z_deltas = data[f"CENTER_Z.{region}"].to_numpy() - data["CENTER_Z"].to_numpy()
    angles = data["angle"].to_numpy() * np.pi / 180.0

    sin_angles = np.sin(angles)
    cos_angles = np.cos(angles)

    return {
        "x": x_deltas * cos_angles - y_deltas * sin_angles,
        "y": x_deltas * sin_angles + y_deltas * cos_angles,
        "z": z_deltas,
    }



from __future__ import annotations

from typing import TYPE_CHECKING

from vtk import vtkPolyData  # pylint: disable=no-name-in-module

from abm_shape_collection.extract_mesh_projections import convert_vtk_to_trimesh

if TYPE_CHECKING:
    import trimesh


def extract_mesh_wireframe(
    mesh: vtkPolyData | trimesh.Trimesh, offset: tuple[float, float, float] | None = None
) -> list[list[tuple[float, float, float]]]:
    """
    Extract wireframe edges from mesh.

    Parameters
    ----------
    mesh
        Mesh object.
    offset
        Mesh translation applied before extracting slices and/or meshes.

    Returns
    -------
    :
        List of wireframe edges.
    """

    if isinstance(mesh, vtkPolyData):
        mesh = convert_vtk_to_trimesh(mesh)

    if offset is not None:
        mesh.apply_translation(offset)

    all_edges = [[tuple(mesh.vertices[a]), tuple(mesh.vertices[b])] for a, b in mesh.edges]
    return [
        [(x1, y1, z1), (x2, y2, z2)]
        for (x1, y1, z1), (x2, y2, z2) in {frozenset(edge) for edge in all_edges}
    ]



from __future__ import annotations

import tempfile
from enum import Enum

import numpy as np
import trimesh
from vtk import vtkPLYWriter, vtkPolyData  # pylint: disable=no-name-in-module

PROJECTIONS: list[tuple[str, tuple[int, int, int], int]] = [
    ("side1", (0, 1, 0), 1),
    ("side2", (1, 0, 0), 0),
    ("top", (0, 0, 1), 2),
]
"""Mesh projection names, normals, and extent axes."""


class ProjectionType(Enum):
    """Projection slice types."""

    SLICE = 1
    """Slice projection type."""

    EXTENT = 2
    """Extent projection type."""


def extract_mesh_projections(
    mesh: vtkPolyData | trimesh.Trimesh,
    projection_types: list[ProjectionType] | None = None,
    offset: tuple[float, float, float] | None = None,
) -> dict:
    """
    Extract slices and/or extents from mesh.

    Slice projections are taken as the cross section of the mesh with planes in
    the x, y, and z directions with origin at (0,0,0). Extent projections are
    taken as cross section of the mesh with planes in the x, y, and z directions
    at increments of 0.5.

    Parameters
    ----------
    mesh
        Mesh object.
    projection_types
        Mesh projection types.
    offset
        Mesh translation applied before extracting slices and/or meshes.

    Returns
    -------
    :
        Map of mesh projection path points.
    """

    if isinstance(mesh, vtkPolyData):
        mesh = convert_vtk_to_trimesh(mesh)

    if projection_types is None:
        projection_types = [ProjectionType.SLICE, ProjectionType.EXTENT]

    if offset is not None:
        mesh.apply_translation(offset)

    projections: dict[str, list[list[list[float]]] | dict[float, list[list[list[float]]]]] = {}

    if ProjectionType.SLICE in projection_types:
        for projection, normal, _ in PROJECTIONS:
            projections[f"{projection}_slice"] = get_mesh_slice(mesh, normal)

    if ProjectionType.EXTENT in projection_types:
        for projection, normal, index in PROJECTIONS:
            projections[f"{projection}_extent"] = get_mesh_extent(mesh, normal, index)

    return projections


def convert_vtk_to_trimesh(mesh: vtkPolyData) -> trimesh.Trimesh:
    """
    Convert VTK polydata to trimesh object.

    Parameters
    ----------
    mesh
        VTK mesh object.

    Returns
    -------
    :
        Trimesh mesh object.
    """
    with tempfile.NamedTemporaryFile() as temp:
        writer = vtkPLYWriter()
        writer.SetInputData(mesh)
        writer.SetFileTypeToASCII()
        writer.SetFileName(f"{temp.name}.ply")
        _ = writer.Write()
        return trimesh.load(f"{temp.name}.ply")


def get_mesh_slice(mesh: trimesh.Trimesh, normal: tuple[int, int, int]) -> list[list[list[float]]]:
    """
    Get slice of mesh along plane for given normal as path points.

    Parameters
    ----------
    mesh
        Mesh object.
    normal
        Vector normal to slice plane.

    Returns
    -------
    :
        List of connected vertices in space specifying the slice.
    """

    mesh_slice = mesh.section_multiplane((0, 0, 0), normal, [0])
    return [[list(point) for point in entity] for entity in mesh_slice[0].discrete]


def get_mesh_extent(
    mesh: trimesh.Trimesh, normal: tuple[int, int, int], index: int
) -> dict[float, list[list[list[float]]]]:
    """
    Get extent of mesh along plane for given normal as path points.

    Parameters
    ----------
    mesh
        Mesh object.
    normal
        Vector normal to slice plane.
    index
        Index of normal axis.

    Returns
    -------
    :
        Map to list of connected vertices in space specifying the extent.
    """

    layers = int(mesh.extents[index] + 2)
    plane_indices = list(np.arange(-layers, layers + 1, 0.5))
    mesh_extents = mesh.section_multiplane((0, 0, 0), normal, plane_indices)
    return {
        index: [[list(point) for point in entity] for entity in mesh_extent.discrete]
        for mesh_extent, index in zip(mesh_extents, plane_indices)
        if mesh_extent is not None
    }



import pandas as pd
from scipy.stats import ks_2samp


def calculate_feature_statistics(
    features: list[str],
    data: pd.DataFrame,
    ref_data: pd.DataFrame,
) -> pd.DataFrame:
    """
    Perform two-sample Kolmogorov-Smirnov test for goodness of fit on features.

    Parameters
    ----------
    features
        List of features to perform test on.
    data
        Sample data, with features as columns.
    ref_data : pd.DataFrame
        References data, with features as columns.

    Returns
    -------
    :
        Kolmogorov-Smirnov statistics and p-values for each feature.
    """

    statistics = []

    for feature in features:
        # Extract values for specific component.
        ref_values = ref_data[feature].to_numpy()
        values = data[feature].to_numpy()

        # Calculate Kolmogorov-Smirnov statistic.
        ks_result = ks_2samp(values, ref_values, mode="asymp")

        statistics.append(
            {
                "FEATURE": feature.upper(),
                "SIZE": len(values),
                "KS_STATISTIC": ks_result.statistic,
                "KS_PVALUE": ks_result.pvalue,
            }
        )

    return pd.DataFrame(statistics)



import numpy as np
import pandas as pd
from aicsshparam import shtools
from vtk import (  # pylint: disable=no-name-in-module
    vtkPolyData,
    vtkTransform,
    vtkTransformPolyDataFilter,
)


def construct_mesh_from_coeffs(
    coeffs: pd.DataFrame,
    order: int,
    prefix: str = "",
    suffix: str = "",
    scale: float = 1.0,
) -> vtkPolyData:
    """
    Construct a mesh from spherical harmonic coefficients.

    Parameters
    ----------
    coeffs
        Spherical harmonic coefficients.
    order
        Order of the spherical harmonics coefficient parametrization.
    prefix
        Prefix string for all coefficient columns.
    suffix
        Suffix string for all coefficient columns.
    scale
        Scale factor for mesh points.

    Returns
    -------
    :
        Mesh object.
    """

    coeffs_map = np.zeros((2, order + 1, order + 1), dtype=np.float32)

    for lix in range(order + 1):
        for mix in range(order + 1):
            coeffs_map[0, lix, mix] = coeffs[f"{prefix}shcoeffs_L{lix}M{mix}C{suffix}"]
            coeffs_map[1, lix, mix] = coeffs[f"{prefix}shcoeffs_L{lix}M{mix}S{suffix}"]

    mesh, _ = shtools.get_reconstruction_from_coeffs(coeffs_map)

    if scale != 1.0:
        transform = vtkTransform()
        transform.Scale((scale, scale, scale))

        transform_filter = vtkTransformPolyDataFilter()
        transform_filter.SetInputData(mesh)
        transform_filter.SetTransform(transform)
        transform_filter.Update()

        mesh = transform_filter.GetOutput(0)

    return mesh



import numpy as np
from skimage.measure import regionprops_table


def get_shape_properties(array: np.ndarray, properties: list[str]) -> dict:
    """
    Calculate shape properties for binary image array.

    Shape properties must be value properties from skimage.measure.regionprops.
    Image is projected along the first axis.

    Parameters
    ----------
    array
        Binary image array.
    properties
        List of shape properties to calculate.

    Returns
    -------
    :
        Dictionary of calculated shape properties.
    """

    flattened_array = array.max(axis=0)

    properties_dict = regionprops_table(flattened_array, properties=properties)
    return {key: value[0] for key, value in properties_dict.items()}



import numpy as np
from aicsshparam import shparam, shtools


def get_shape_coefficients(array: np.ndarray, reference: np.ndarray, order: int) -> dict:
    """
    Calculate spherical harmonic coefficients for binary image array.

    Parameters
    ----------
    array
        Binary image array.
    reference
        Binary reference array that determines alignment angle.
    order
        Order of the spherical harmonics coefficient parametrization.

    Returns
    -------
    :
        Dictionary of spherical harmonics, angle, and MSE.
    """

    _, angle = shtools.align_image_2d(image=reference)
    aligned_array = shtools.apply_image_alignment_2d(array, angle).squeeze()

    (coeffs, reconstructed_array), (_, _, downsampled_array, _) = shparam.get_shcoeffs(
        image=aligned_array, lmax=order, compute_lcc=False, alignment_2d=False
    )

    mse = shtools.get_reconstruction_error(downsampled_array, reconstructed_array)
    coeffs["angle"] = angle
    coeffs["mse"] = mse

    return coeffs



"""Collection of tasks for analyzing cell shapes."""

from prefect import task

from .calculate_feature_statistics import calculate_feature_statistics
from .calculate_shape_statistics import calculate_shape_statistics
from .construct_mesh_from_array import construct_mesh_from_array
from .construct_mesh_from_coeffs import construct_mesh_from_coeffs
from .construct_mesh_from_points import construct_mesh_from_points
from .extract_mesh_projections import extract_mesh_projections
from .extract_mesh_wireframe import extract_mesh_wireframe
from .extract_shape_modes import extract_shape_modes
from .extract_voxel_contours import extract_voxel_contours
from .fit_pca_model import fit_pca_model
from .get_shape_coefficients import get_shape_coefficients
from .get_shape_properties import get_shape_properties
from .make_voxels_array import make_voxels_array

calculate_shape_statistics = task(calculate_shape_statistics)
calculate_feature_statistics = task(calculate_feature_statistics)
construct_mesh_from_array = task(construct_mesh_from_array)
construct_mesh_from_coeffs = task(construct_mesh_from_coeffs)
construct_mesh_from_points = task(construct_mesh_from_points)
extract_mesh_projections = task(extract_mesh_projections)
extract_mesh_wireframe = task(extract_mesh_wireframe)
extract_shape_modes = task(extract_shape_modes)
extract_voxel_contours = task(extract_voxel_contours)
fit_pca_model = task(fit_pca_model)
get_shape_coefficients = task(get_shape_coefficients)
get_shape_properties = task(get_shape_properties)
make_voxels_array = task(make_voxels_array)
